/*!\file
\brief The main file of watercolor2rgba.

A command-line tool that equips scanned images with transparency 
information.

watercolor2rgba is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 2 of the License, or
(at your option) any later version.

watercolor2rgba is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#define SAMPLE_MAXVALUE 255

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <getopt.h>
#include <Imlib2.h>
#include <ctype.h>
#include "gettext.h"
#define _(String) gettext (String)
#include "configure.h"

/*! The help message this program can output

Originally I did try to put this message into a constant in the main
function: If the task is more complicated than this one I use C++ and
having global variables is a thing you learn to avoid there. But gettext
didn't seem to like filling the constant with a translation afterwards.

Now we have a pointer to the help message instead - and an atexit function
that frees the memory area it points to on exit.
 */
char* helpmessage;

/*! \brief This function is called on exit of the main program.

Didn't find another convenient way of freeing the help message on exit.
Theoretically we don't have to do it since linux automatically does this
job. But you never know which operating system your software will run
in the end.
*/
void exitfunc()
{
  free(helpmessage);
}

/* Is this pixel in a transparent region?
   
   \param x The x-position of the pixel
   \param y The y-position of the pixel
   \param width  The width of the image
   \param height The hight of the image
   \param threshold The minimum transparency that is deemed to be transparent
   \param WhiteBackground A pointer to the image scanned with a white background
   \param BlackBackground A pointer to the image scanned with a black background
 */
int pixel_is_transparent(long int x,long int y,
			 long int width,long int height,
			 int threshold,
			 Imlib_Image WhiteBackground,
			 Imlib_Image BlackBackgnd)
{
  int retval=1;
  long int i,o;
  long int xmin=x-2,xmax=x+2,ymin=y-2,ymax=y+2;

  Imlib_Color Pixel_WhiteBackgnd[5][5];
  Imlib_Color Pixel_BlackBackgnd[5][5];


  if(xmin<0)
    xmin=0;
  if(ymin<0)
    ymin=0;
  if(xmax>width)
    xmax=width;
  if(ymax>height)
    ymax=height;

  imlib_context_set_image(BlackBackgnd);
  for(i=xmin;i<=xmax;i++)
    for(o=ymin;o<=ymax;o++)
      imlib_image_query_pixel(i,o,&Pixel_BlackBackgnd[i-xmin][o-ymin]);
  
  imlib_context_set_image(WhiteBackground);
  for(i=xmin;i<=xmax;i++)
    for(o=ymin;o<=ymax;o++)
      imlib_image_query_pixel(i,o,&Pixel_WhiteBackgnd[i-xmin][o-ymin]);

  for(i=xmin;i<=xmax;i++)
    for(o=ymin;o<=ymax;o++)
      {
	int difference;	  
	  difference=Pixel_WhiteBackgnd[i-xmin][o-ymin].red -Pixel_BlackBackgnd[i-xmin][o-ymin].red+
	    Pixel_WhiteBackgnd[i-xmin][o-ymin].green-Pixel_BlackBackgnd[i-xmin][o-ymin].green+
	    Pixel_WhiteBackgnd[i-xmin][o-ymin].blue -Pixel_BlackBackgnd[i-xmin][o-ymin].blue;

	  if(difference<=threshold) 
	    retval=0;
	  //	  printf("Diff %i Th %i\n",difference,threshold);
	}
  return(retval);
}


/*! The main function */
int main(int argc, char **argv)
{
  //! How many percents above the autodetected value do we deem to be "transparent"?
  int mintransparencyoffset=10;
  //! How many percents below the autodetected value do we deem to be "opaque"? 
  int maxtransparencyoffset=10;
  //! The image scanned with a white background layn behind it
  Imlib_Image WhiteBackground;
  //! The image scanned with a black background layn behind it
  Imlib_Image BlackBackgnd;
  //! The output image
  Imlib_Image Dest;
  //! The color of the paper itself
  Imlib_Color Papercolor;
  //! The percentage of thrasparency that is valued "fully transparent"
  int TransparencyCutOffPercentage=100;

  helpmessage = strdup(_("Compares two version of the same image:\n"
      "One version scanned with a white sheet of paper behind it\n"
      "and one scanned with a black sheet of paper put behind it instead:\n"
      "The result will be the original image with added transparency information\n\n"
      "Usage:\n\n"
			 "watercolor2rgba InputfileScannedWithWhiteBackground.png InputfileScannedWithBlackBackground.png Outputfile.png\n"));
 
  atexit(exitfunc);

  // The following block interprets our command-line arguments.
  {
    static struct option long_options[] =
      {
	{"help",     no_argument,       0, 'h'},
	{"version",  no_argument,       0, 'v'},
	{"min-threshold-offset",     required_argument,       0, 'm'},
	{"max-threshold-offset",     required_argument,       0, 'M'},
	{"transparency-cutoff",      required_argument,       0, 'C'},
	{0, 0, 0, 0}
      };
    
    int index;
    int c;
    
    opterr = 0;
    
    while ((c = getopt_long (argc, argv, "hvm:M:C:",long_options,&index)) != -1)
      switch (c)
	{
	case'v':
	  fprintf(stderr,_("%s v. %s\n"),argv[0],PACKAGE_VERSION);
	  fprintf(stderr,"Copyright (C) 2010-2011 Gunter Koenigsmann.\n");
	  fprintf(stderr,"License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>.\n");
	  fprintf(stderr,"This is free software: you are free to change and redistribute it.\n");
	  fprintf(stderr,"There is NO WARRANTY, to the extent permitted by law.\n");
	  exit(0);
	  break;
	case 'h':
	  {
	    fprintf (stderr, "%s",helpmessage);
	    exit(0);
	  }
	  break;
	case 'm':
	  errno=0;
	  mintransparencyoffset=strtol(optarg,NULL,10);
	  if(errno)
	    {
	      fprintf(stderr,_("Cannot interpret the min transparency offset factor\n"));
	      exit(-1);
	    }
	  else
	    {
	      printf(_("Setting the offset to the min transparency detection to %i%%\n"),maxtransparencyoffset);
	    }
	  break;
	case 'M':
	  errno=0;
	  maxtransparencyoffset=strtol(optarg,NULL,10);
	  if(errno)
	    {
	      fprintf(stderr,_("Cannot interpret the max transparency offset factor\n"));
	      exit(-1);
	    }
	  else
	    {
	      printf(_("Setting the offset to the max transparency detection to %i%%\n"),maxtransparencyoffset);
	    }
	  break;
	case 'C':
	  errno=0;
	  TransparencyCutOffPercentage=strtol(optarg,NULL,10);
	  if(errno)
	    {
	      fprintf(stderr,_("Cannot interpret the tansparency cutoff percentage %s\n"),optarg);
	      exit(-1);
	    }
	  else
	    {
	      printf(_("Pixels whose transparency is > %i%% are now interpreted as 100%% transparent\n"),TransparencyCutOffPercentage);
	    }
	  break;

	case '?':
	  if (optopt == 'c')
	    {
	      fprintf (stderr, _("Option -%c requires an argument.\n"), optopt);
	      exit(-1);
	    }
	  else if (isprint (optopt))
	    {
	      fprintf (stderr, _("Unknown option `-%c'.\n"), optopt);
	      exit(-1);
	    }
	  else
	    {
	      fprintf (stderr,
		       _("Unknown option character `\\x%x'.\n"),
		       optopt);
	      exit(-1);
	    }
	  return 1;
	default:
	  abort ();
	}
  }

  
  /* We need 3 filenames as arguments.
     If we don't get them this is an error.
  */
  if (argc - optind != 3)
    {
      fprintf(stderr,"%s",helpmessage);
      exit(1);
    }

  printf(_("Loading the image scanned with a white background\n"));
  if((WhiteBackground=imlib_load_image(argv[optind]))!=0)
    {      
      printf(_("Loading the image scanned with a black background\n"));
      if((BlackBackgnd=imlib_load_image(argv[optind+1]))!=0)
	{
	  // The image dimensions
	  long int width,height;
	  // Counters for for loops
	  long int x,y;
	  int i;
	  // The transparency values we detected
	  int min_transparency=0,max_transparency=0;

	  long int papercolor_red=0,papercolor_green=0,papercolor_blue=0;

	  // The transparency histogram
	  long int histogram[3*SAMPLE_MAXVALUE];
	  // Theoretically there is no negative transparency. But if the 
	  // images aren't aligned correctly or the scanning process cannot 
	  // be reproduced this accurately we will see one.
	  int min_negative_transparency=0;

	  // Detect the size of the input images
	  imlib_context_set_image(WhiteBackground);
	  width=imlib_image_get_width();
	  height=imlib_image_get_height();
	  
	  // Paranoia: What if both images differ in size?
	  imlib_context_set_image(BlackBackgnd);
	  if(width!=imlib_image_get_width()||height!=imlib_image_get_height())
	    {
	      fprintf(stderr,_("Images differ in size! Aborting.\n"));
	      exit(-1);
	    }
	  
	  // Create the destination image and fill it with transparency.
	  Dest=imlib_create_image(width,height);
	  imlib_context_set_image(Dest);
	  imlib_image_set_has_alpha(1);
	  imlib_image_clear_color(0,0,0,0);

	  // Generate a transparency histogram
	  {
	    printf(_("Calculating a transparency histogram of the image\n"));

	    memset(&histogram,0,sizeof(histogram));
	    
	    for(y=0;y<width;y++)
	      for(x=0;x<width;x++)
		{
		  int difference;
		  Imlib_Color Pixel_WhiteBackgnd;
		  Imlib_Color Pixel_BlackBackgnd;
		  imlib_context_set_image(WhiteBackground);
		  imlib_image_query_pixel(x,y,&Pixel_WhiteBackgnd);
		  imlib_context_set_image(BlackBackgnd);
		  imlib_image_query_pixel(x,y,&Pixel_BlackBackgnd);

		  difference=Pixel_WhiteBackgnd.red  -Pixel_BlackBackgnd.red+
		    Pixel_WhiteBackgnd.green-Pixel_BlackBackgnd.green+
		    Pixel_WhiteBackgnd.blue -Pixel_BlackBackgnd.blue;
		  if((difference>=0)&&(difference<(3*SAMPLE_MAXVALUE)))
		    histogram[difference]++;

		  if(difference<min_negative_transparency)
		    min_negative_transparency=difference;
		}

	    if(min_negative_transparency<-20)
	      printf(_("Warning: Alignment of the images might be bad: Minimum Transparency=%i\n"),
		     min_negative_transparency);
	  }
	  
	  // Search for the most transparent 3% of the image: We lateron
	  // will assume that they are of the paper's color. 
	  {
	    int paper_transparency;
	    long int pixels_of_paper_color=0;
	    paper_transparency=3*SAMPLE_MAXVALUE-6;
	    while(
		  (pixels_of_paper_color<width*height*3/100)&&
		  (paper_transparency>0)
		  )
	      {
		pixels_of_paper_color+=histogram[paper_transparency];
		paper_transparency--;
	      }
	    printf(_("Assuming pixels with a transparency >%i to be of the paper's color.\n"),paper_transparency);
	    
	    printf(_("Now calculating the paper color, which might take a while.\n"));
	    long int papercolor_pixels=0;
	    for(y=0;y<width;y++)
	      for(x=0;x<width;x++)
		{
		  if(pixel_is_transparent(x,y,width,height,paper_transparency,WhiteBackground,BlackBackgnd))
		    {
		      Imlib_Color Pixel;
		      imlib_context_set_image(WhiteBackground);
		      imlib_image_query_pixel(x,y,&Pixel);
		      papercolor_red+=Pixel.red;
		      papercolor_green+=Pixel.green;
		      papercolor_blue+=Pixel.blue;
		      papercolor_pixels++;
		    }
		}
	    if(papercolor_pixels)
	      {
		papercolor_red/=papercolor_pixels;
		papercolor_green/=papercolor_pixels;
		papercolor_blue/=papercolor_pixels;
	      } else
	      {
		printf(_("Unable to determine the color of the paper.\n"
			 "Using plain white instead.\n"));
		papercolor_red=SAMPLE_MAXVALUE;
		papercolor_green=SAMPLE_MAXVALUE;
		papercolor_blue=SAMPLE_MAXVALUE;
		
	      }
	    printf(_("Paper color: #%02x%02x%02x\n"),Papercolor.red=papercolor_red,Papercolor.green=papercolor_green,Papercolor.blue=papercolor_blue);
	  }

	  // Try to read out the transparency span
	  // TODO: for > 8 bits per sample we will need a higher filter constant
	  //       here and for 16 Bits per sample either an IIR filter (which
	  //       is faster than this FIR one). 
	  {
	    printf(_("Trying to find the transparency span: "));
	    long int peaksteepness_last=0;
	    
	    i=3;
	    
	    // Sometimes the derivation of the transparency curve begins with 
	    // a negative value. Let's skip this part.
	    while(((peaksteepness_last=-histogram[i-1]-histogram[i-2]-histogram[i-3]+
		    histogram[i+1]+histogram[i+2]+histogram[i+3])<0)&&(i<3*SAMPLE_MAXVALUE-11))
	      i++;

	    // Now let's search for the first and the last peak of the
	    // transparency histogram.
	    for(;i<3*SAMPLE_MAXVALUE-11;i++)
	      {
		int peaksteepness=-histogram[i-1]-histogram[i-2]-histogram[i-3]+
		  histogram[i+1]+histogram[i+2]+histogram[i+3];
		if((peaksteepness<0)&&(peaksteepness_last>0))
		  {
		    if(min_transparency==0) min_transparency=i;
		    max_transparency=i;
		  }
		peaksteepness_last = peaksteepness;
	      }
	    
	    // Allow some noise margin for "fully transparent" and "fully oblique".
	    min_transparency+=(max_transparency-min_transparency)*mintransparencyoffset/100;
	    max_transparency-=(max_transparency-min_transparency)*maxtransparencyoffset/100;
	    
	    if(min_transparency<0)min_transparency=0;
	    if(max_transparency>=(3*SAMPLE_MAXVALUE))max_transparency=3*SAMPLE_MAXVALUE;

	    printf(_("Min=%i, Max=%i\n"),
		   min_transparency,max_transparency);
	    
	    if (max_transparency<=min_transparency)
	      {
		fprintf(stderr,_("Didn't succeed analyzing the transparency histogram.\n"));
		exit(-1);
	      }
	    
	    max_transparency = max_transparency-min_transparency;
	  }

	  // Generate the output image.
	  {
	    printf(_("Now calculating the image.\n"));
	    for(y=0;y<height;y++)
	      for(x=0;x<width;x++)
		{
		  int difference;
		  Imlib_Color Pixel_WhiteBackgnd;
		  Imlib_Color Pixel_BlackBackgnd;
		  int red,green,blue,alpha;
		  Imlib_Color NewPixel;
		  imlib_context_set_image(BlackBackgnd);
		  imlib_image_query_pixel(x,y,&Pixel_BlackBackgnd);
		  imlib_context_set_image(WhiteBackground);
		  imlib_image_query_pixel(x,y,&Pixel_WhiteBackgnd);

		  // Calculate how big a difference the background makes.
		  difference=Pixel_WhiteBackgnd.red  -Pixel_BlackBackgnd.red+
		    Pixel_WhiteBackgnd.green-Pixel_BlackBackgnd.green+
		    Pixel_WhiteBackgnd.blue -Pixel_BlackBackgnd.blue-min_transparency;

		  
		  if(difference>max_transparency)difference=max_transparency;

		  if(difference<min_transparency)difference=min_transparency;

		  alpha=SAMPLE_MAXVALUE*(max_transparency-difference)/(max_transparency-min_transparency);
		  // Transparency cutoff
		  if(alpha>TransparencyCutOffPercentage*SAMPLE_MAXVALUE/100)
		    alpha=SAMPLE_MAXVALUE;

		  red=  (Pixel_WhiteBackgnd.red+(SAMPLE_MAXVALUE-Papercolor.red)*(SAMPLE_MAXVALUE-alpha)/SAMPLE_MAXVALUE)*(alpha)/SAMPLE_MAXVALUE;
		  green=(Pixel_WhiteBackgnd.green+(SAMPLE_MAXVALUE-Papercolor.red)*(SAMPLE_MAXVALUE-alpha)/SAMPLE_MAXVALUE)*(alpha)/SAMPLE_MAXVALUE;
		  blue= (Pixel_WhiteBackgnd.blue+(SAMPLE_MAXVALUE-Papercolor.red)*(SAMPLE_MAXVALUE-alpha)/SAMPLE_MAXVALUE) *(alpha)/SAMPLE_MAXVALUE;
			 
		  if(red>SAMPLE_MAXVALUE)red=SAMPLE_MAXVALUE;
		  if(green>SAMPLE_MAXVALUE)green=SAMPLE_MAXVALUE;
		  if(blue>SAMPLE_MAXVALUE)blue=SAMPLE_MAXVALUE;

		  NewPixel.red=red;
		  NewPixel.green=green;
		  NewPixel.blue=blue;
		  NewPixel.alpha=alpha;
		  imlib_context_set_color(NewPixel.red,NewPixel.green,NewPixel.blue,NewPixel.alpha);
		  imlib_context_set_image(Dest);
		  imlib_image_draw_pixel(x,y,1);
		}
	  }
	  
	  // It is good practice to always free all images we don't need.
	  imlib_context_set_image(BlackBackgnd);
	  imlib_free_image();
	  imlib_context_set_image(WhiteBackground);
	  imlib_free_image();

	  // Save the image
	  printf(_("Saving the image.\n"));
	  imlib_context_set_image(Dest);
	  char *tmp;
	  /* argument - i.e. .png = png, .tif = tiff etc. */
	  tmp = strrchr(argv[optind+2], '.');
	  if(tmp)
	    imlib_image_set_format(tmp + 1);
	  /* save the image */
	  imlib_save_image(argv[optind+2]);
	  imlib_free_image();
	}
      else
	{
	  fprintf(stderr,_("Error loading image %s!\n"),argv[1]);
	  exit(-1);
	}
    }
  else
    {
      fprintf(stderr,_("Error loading image %s!\n"),argv[2]);
      exit(-1);
    }
  exit(0);
}
